package se.hig.wasteland.evolutionary;
//=============================================================================
// Copyright 2006-2010 Daniel W. Dyer
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//=============================================================================


import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.List;

import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JTextArea;
import javax.swing.SwingUtilities;

import org.uncommons.maths.random.MersenneTwisterRNG;
import org.uncommons.maths.random.Probability;
import org.uncommons.watchmaker.framework.EvolutionObserver;
import org.uncommons.watchmaker.framework.EvolutionaryOperator;
import org.uncommons.watchmaker.framework.GenerationalEvolutionEngine;
import org.uncommons.watchmaker.framework.PopulationData;
import org.uncommons.watchmaker.framework.factories.StringFactory;
import org.uncommons.watchmaker.framework.interactive.Renderer;
import org.uncommons.watchmaker.framework.operators.StringCrossover;
import org.uncommons.watchmaker.framework.operators.StringMutation;
import org.uncommons.watchmaker.framework.operators.EvolutionPipeline;
import org.uncommons.watchmaker.framework.selection.RouletteWheelSelection;
import org.uncommons.watchmaker.framework.termination.GenerationCount;
import org.uncommons.watchmaker.swing.evolutionmonitor.EvolutionMonitor;

import se.hig.jdalmas.norms.Ground;

/**
 * Simple evolutionary algorithm that evolves a population of randomly-generated
 * strings until at least one matches a specified target string.
 * @author Daniel Dyer
 */
public final class StringBasedEvolutionaryWasteland
{
    private static final char[] ALPHABET = new char[9];
    static
    {
        for (char c = '1'; c <= '9'; c++)
        {
            ALPHABET[c - '1'] = c;
        }
    }

    private static final int PORT = 50009;
    private static StringBasedWastecollectorsEvaluator evaluator;
	private static EvolutionMonitor<String> monitor;
//	private static final int N_EVENTS = 49;
	private static final int N_IND = 100;
	private static final double ELITISM = 0.25;
	private static final int N_GEN = 5;
	private static final int N_GROUNDS = StringBasedWastecollectorsEvaluator.N_GROUNDS;
	private static final int N_CONS = StringBasedWastecollectorsEvaluator.N_CONS;
	private static final int N_CROSS = N_GROUNDS-1;
	private static final Probability crossoverProbability = new Probability(0.7d);
	private static final Probability mutationProbability = new Probability(0.05d);

	
   public static void main(String[] args)
    {
        String result;
		try {
			result = evolveString();
	        System.out.println("Evolution result:\n" + evaluator.toString(result,","));
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
    }


    public static String evolveString() throws IOException
    {
        List<EvolutionaryOperator<String>> operators = new ArrayList<EvolutionaryOperator<String>>(2);
        operators.add(new StringCrossover(N_CROSS,crossoverProbability));
        operators.add(new StringMutation(ALPHABET,mutationProbability));
        StringFactory factory = new StringFactory(ALPHABET,N_CONS);
        EvolutionaryOperator<String> pipeline = new EvolutionPipeline<String>(operators);
        evaluator = new StringBasedWastecollectorsEvaluator(PORT,ALPHABET);
        String bestCandidate = null;
        GenerationalEvolutionEngine<String> engine = new GenerationalEvolutionEngine<String>(
        		factory,
                pipeline,
                evaluator,
                new RouletteWheelSelection(),
                new MersenneTwisterRNG());
        FileWriter writer = new FileWriter(createFileName());
        writeLogPreamble(writer);
        writeLogHeadings(writer);
        engine.addEvolutionObserver(new EvolutionLogger(writer));
        Renderer<String,JComponent> renderer = new Renderer<String,JComponent>() {
        	public JComponent render(String candidate) {
        		return new JTextArea(StringBasedWastecollectorsEvaluator.toString(candidate,"\n"));
        	}
        };
        monitor = new EvolutionMonitor<String>(renderer,false);
        engine.addEvolutionObserver(monitor);
        SwingUtilities.invokeLater(new Runnable() {

			@Override
			public void run() {
				JFrame frame = new JFrame();
				frame.add(monitor.getGUIComponent());
				frame.pack();
				frame.setVisible(true);
				frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
			}
        	
        });
        engine.setSingleThreaded(true);
        int elitism = (int)Math.round(N_IND*ELITISM);
        System.out.println("Elitism: " + elitism);
        bestCandidate = engine.evolve(N_IND, // N_IND individuals in the population.
        		elitism, // Level of elitism.
        		factory.generateInitialPopulation(N_IND,createSeeds(N_IND/2),new MersenneTwisterRNG()),
                new GenerationCount(N_GEN));
        writer.close();
        return bestCandidate;
    }

	private static void writeLogPreamble(FileWriter writer) throws IOException {
		writer.write("Population size:\t" + N_IND + "\n");
		writer.write("Elitism:\t" + ELITISM + "\n");
		writer.write("Crossover probability:\t" + crossoverProbability + "\n");
		writer.write("Crossover points:\t" + N_CROSS + "\n");
		writer.write("Mutation probability:\t" + mutationProbability + "\n");
		writer.write("Grounds and norms:\n");
        for(int i=0; i<StringBasedWastecollectorsEvaluator.grounds.length; i++) {
        	writer.write(StringBasedWastecollectorsEvaluator.grounds[i] + "(" + 
        			Arrays.toString(StringBasedWastecollectorsEvaluator.groundParams[i]) + "): ");
        	for(int j=0; j<StringBasedWastecollectorsEvaluator.consequenceParams[i].length; j++) {
        		writer.write(Arrays.toString(StringBasedWastecollectorsEvaluator.consequenceParams[i][j])+",");
        	}
        	writer.write("\n");
        }
	}

    private static void writeLogHeadings(FileWriter writer) throws IOException {
        writer.write("\nGen\tFitness\tMean\tSD\tBest candidate\t");
        for(int i=0; i<N_GROUNDS; i++)
        	for(int j=0; j<StringBasedWastecollectorsEvaluator.consequenceParams[i].length; j++)
        		writer.write("n"+i+""+j+"\t");
        writer.write("\n");
        writer.flush();
	}


	private static String createFileName() {
    	Calendar now = Calendar.getInstance();
		return "logs/Log " + now.get(Calendar.YEAR) + "-" +
    			to2DigitString(now.get(Calendar.MONTH)+1) + "-" +
    			to2DigitString(now.get(Calendar.DATE)) + " " + 
    			to2DigitString(now.get(Calendar.HOUR_OF_DAY)) + "." +
    			to2DigitString(now.get(Calendar.MINUTE)) + ".txt";
	}


	private static String to2DigitString(int i) {
		return (i<10?"0"+i:""+i);
	}


	private static List<String> createSeeds(int n) {
        List<String> seeds = new ArrayList<>(n);
        for(int i=0; i<n; i++) {
        	StringBuffer sBuf = new StringBuffer(N_CONS);
        	for(int j=0; j<N_CONS; j++)
        		if(i%2==0 && (j==5))
        			sBuf.append('9');
        		else
        			sBuf.append('1');
        	seeds.add(sBuf.toString());
        }
        System.out.println(seeds);
		return seeds;
	}

	/**
     * Trivial evolution observer for displaying information at the end
     * of each generation.
     */
    private static class EvolutionLogger implements EvolutionObserver<String>
    {
    	
        private FileWriter writer;

		public EvolutionLogger(FileWriter writer) {
        	super();
        	this.writer = writer;
		}

		public void populationUpdate(PopulationData<? extends String> data)
        {
            try {
            	int gen = data.getGenerationNumber();
            	System.err.println("Generation: " + gen +"\n");
				writer.write(gen + "\t" + 
						data.getBestCandidateFitness() + "\t" +
				        data.getMeanFitness() + "\t" +
				        data.getFitnessStandardDeviation() + "\t" +
				        StringBasedWastecollectorsEvaluator.toString(data.getBestCandidate(),"\t") + 
				        "\n");
				writer.flush();
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
        }
    }
    
}
