//=============================================================================
// Copyright 2006-2010 Daniel W. Dyer
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//=============================================================================
package se.hig.wasteland.evolutionary;

import java.util.Arrays;
import java.util.List;

import javax.management.RuntimeErrorException;

import org.uncommons.maths.binary.BitString;
import org.uncommons.watchmaker.framework.FitnessEvaluator;

import se.hig.jdalmas.Predicate;
import se.hig.jdalmas.control.DnrDALMASController;
import se.hig.jdalmas.norms.Consequence;
import se.hig.jdalmas.norms.ElementaryConsequence;
import se.hig.jdalmas.norms.ElementaryGround;
import se.hig.jdalmas.norms.ElementaryNorm;
import se.hig.jdalmas.norms.Ground;
import se.hig.jdalmas.norms.MoveOperator;
import se.hig.jdalmas.norms.Norm;
import se.hig.jdalmas.norms.NormSystem;
import se.hig.jdalmas.norms.Operator;
import se.hig.jdalmas.norms.SitCondition;
import se.hig.jdalmas.norms.StateCondition;

/**
 * A fitness evaluator that simply counts the number of ones in a bit
 * string.
 * @see BitString
 * @author Daniel Dyer
 */
public class StringBasedWastecollectorsEvaluator implements FitnessEvaluator<String>
{
//	private static final BigInteger B63 = BigInteger.valueOf(63);
	private static final String[] ZERO = {"0"};
	private static final String[] ONE = {"1"};
	private static final String[] TWO = {"2"};
	private static final String[] THREE = {"3"};
	private static final String[] FOUR = {"4"};
	private static final String[] SIX = {"6"};
	private static final String[] NINE = {"9"};
	private static final String LAP = "lap";
	private static final String DIFF_AND_LAP = "diff_and_lap";
	private static final String PASS = "pass";
	private static final String USER = "user";
	private static final String GO = "go";

	private int port;
   	private WasteDALMASController wasteDALMASController;
	private char[] alphabet;
	public static final String[] grounds = {DIFF_AND_LAP,LAP,LAP,LAP,LAP,LAP,LAP};
	public static final String[][] groundParams = {NINE,SIX,FOUR,THREE,TWO,ONE,ZERO};
	public static final int N_GROUNDS = groundParams.length;
	public static final String[][][] consequenceParams = { // Movement in four directions
		{NINE}, // M1(diff&lap9)
		{NINE,SIX,FOUR,THREE}, // M1lap6
		{SIX,FOUR,TWO}, // M1lap4
		{SIX,THREE,TWO,ZERO}, // M1lap3
		{FOUR,THREE,TWO,ONE,ZERO}, // M1lap2
		{TWO,ONE,ZERO}, // M1lap1
		{THREE,TWO,ONE,ZERO}, // M1lap0		
		};
//	public static final String[][][] consequenceParams = { // Movement in eight directions
//		{NINE,SIX,FOUR}, // M1(diff&lap9)
//		{NINE,SIX,FOUR,THREE,TWO}, // M1lap6
//		{NINE,SIX,FOUR,THREE,TWO,ONE}, // M1lap4
//		{SIX,FOUR,THREE,TWO,ZERO}, // M1lap3
//		{SIX,FOUR,THREE,TWO,ONE,ZERO}, // M1lap2
//		{FOUR,TWO,ONE,ZERO}, // M1lap1
//		{THREE,TWO,ONE,ZERO}, // M1lap0		
//		};
	public static final int N_CONS = 
			consequenceParams[0].length + 
			consequenceParams[1].length + 
			consequenceParams[2].length + 
			consequenceParams[3].length + 
			consequenceParams[4].length + 
			consequenceParams[5].length + 
			consequenceParams[6].length;
	private static final int[] consequenceType = {1,21,22,41,42,5,61,62,7};

	
    public StringBasedWastecollectorsEvaluator(int port, char [] alphabet) {
//      public StringBasedWastecollectorsEvaluator(int port, char [] alphabet, int nEvents) {
  	this.port = port;
  	this.alphabet = alphabet;
//  	this.nEvents = nEvents;
	}

    
	private void initDnrDALMAS(int port, int rows, int cols) {
		wasteDALMASController = new WasteDALMASController(new DnrDALMASController("localhost",port,"_"));
		wasteDALMASController.getDnrDALMASController().addAgent("x1");
		wasteDALMASController.getDnrDALMASController().addAgent("x2");
		wasteDALMASController.getDnrDALMASController().addAgent("x3");
//		wasteDALMASController.getDnrDALMASController().addAgent("x4");
		wasteDALMASController.addAct(PASS);
		wasteDALMASController.addAct(GO,"east");
//		wasteDALMASController.addAct(GO,"northeast");
		wasteDALMASController.addAct(GO,"north");
//		wasteDALMASController.addAct(GO,"nortwest");
		wasteDALMASController.addAct(GO,"west");
//		wasteDALMASController.addAct(GO,"southwest");
		wasteDALMASController.addAct(GO,"south");
//		wasteDALMASController.addAct(GO,"southeast");
		wasteDALMASController.getDnrDALMASController().addPredicate(Predicate.FEASIBLE_ACT_PREDICATE, 
				USER,"agentFeasibleAct",3);
		wasteDALMASController.getDnrDALMASController().addPredicate(Predicate.UTILITY_FUNCTION_PREDICATE,
				USER,"agentUtilityFunction",4);
		wasteDALMASController.getDnrDALMASController().addPredicate(Predicate.TIE_BREAK_PREDICATE,
				USER,"tieBreaker_random",4);
		wasteDALMASController.addBoundaries(0,0,rows+1,cols+1);
		wasteDALMASController.addFixedWaste();
		wasteDALMASController.addAgentPosition("x1",1,1);
		wasteDALMASController.addAgentPosition("x2",1,1);
		wasteDALMASController.addAgentPosition("x3",1,1);
//		wasteDALMASController.addAgentPosition("x4",1,1);
		wasteDALMASController.addAgentWasteLoad("x1",0);
		wasteDALMASController.addAgentWasteLoad("x2",0);
		wasteDALMASController.addAgentWasteLoad("x3",0);
//		wasteDALMASController.addAgentWasteLoad("x4",0);
		wasteDALMASController.addPreviousAct("x1",PASS,null);
		wasteDALMASController.addPreviousAct("x2",PASS,null);
		wasteDALMASController.addPreviousAct("x3",PASS,null);
//		wasteDALMASController.addPreviousAct("x4",PASS,null);
		wasteDALMASController.getDnrDALMASController().setAgentToMove("x1");
	}
	

	/**
     * Calculates a fitness score for the candidate string.
     * @param candidate The evolved string to evaluate.
     * @param population {@inheritDoc}
     * @return Mean of evaluation score.
     */
//    public double getFitness(String candidate,
//                             List<? extends String> population)
//    {
//    	initDnrDALMAS(port);
//    	addNormativeSystem(candidate);
//    	wasteDALMASController.getDnrDALMASController().kEventRun(nEvents);
//        return evaluateSituation();
//    }    
    public double getFitness(String candidate,
    		List<? extends String> population)
    {
    	int rows, cols;
    	NormSystem normSystem = createNormativeSystem(candidate);
    	double sum=0.0;
    	// 1st run: 6 x 8
    	rows = 6; cols = 8;
    	initDnrDALMAS(port,rows,cols);
    	wasteDALMASController.getDnrDALMASController().getCurrentSituation().getState().getDnrDALMAS().setNormSystem(normSystem);
    	wasteDALMASController.getDnrDALMASController().kEventRun(rows*cols-1);
    	sum += evaluateSituation();
    	// 2nd run: 7 x 7
    	rows = cols = 7;
    	initDnrDALMAS(port,rows,cols);
    	wasteDALMASController.getDnrDALMASController().getCurrentSituation().getState().getDnrDALMAS().setNormSystem(normSystem);
    	wasteDALMASController.getDnrDALMASController().kEventRun(rows*cols-1);
    	sum += evaluateSituation();
    	// 3rd run: 10 x 5
    	rows = 10; cols = 5;
    	initDnrDALMAS(port,rows,cols);
    	wasteDALMASController.getDnrDALMASController().getCurrentSituation().getState().getDnrDALMAS().setNormSystem(normSystem);
    	wasteDALMASController.getDnrDALMASController().kEventRun(rows*cols-1);
    	sum += evaluateSituation();
    	
    	return sum/3;
    }


    private NormSystem createNormativeSystem(String candidate) {
    	NormSystem normSystem = NormSystem.createEmptyNormSystem(candidate);
    	// GROUNDS
		StateCondition lap = StateCondition.createStateCondition(USER,"overlap",1,2);
		normSystem.addGroundToGroundSet(LAP,lap);
		StateCondition diff_and_lap = StateCondition.createStateCondition(USER,DIFF_AND_LAP,1,2);
		normSystem.addGroundToGroundSet(DIFF_AND_LAP,diff_and_lap);
		// CONSEQUENCES
		normSystem.addConsequenceToConsequenceSet(LAP,lap);
		// NORMS
		System.out.print("Norms: ");
    	for(int i=0,pos=0; i<N_GROUNDS; i++) {
    		ElementaryGround ground = Ground.createElementaryGround(
    				MoveOperator.M[1],grounds[i],groundParams[i],normSystem.getGroundSet());
    		String [][] cons = consequenceParams[i];
    		for(int j=0; j<cons.length; j++) {
    			Operator consequenceOp = Operator.createOperator(
    					Operator.TYPE,getType(candidate.charAt(pos))+"");
        		ElementaryConsequence consequence = Consequence.createElementaryConsequence(
        				consequenceOp,LAP,cons[j],normSystem.getConsequenceSet());
    			Norm norm = NormSystem.createElementaryNorm(
    					candidate,"n"+i+""+j,ground,consequence);
    			System.out.print(norm + ",");
    			normSystem.addNorm(norm);
//    			wasteDALMASController.getDnrDALMASController().addElementaryNorm(
//    					"n"+i+j,1,grounds[i],groundParams[i],
//    					getType(candidate.charAt(pos)),
//    					LAP,cons[j]
//    					);
    			pos++;
    		}
    	}
		System.out.println();
		return normSystem;
    }

    public static String toString(String candidate, String separator) {
    	StringBuffer sBuf = new StringBuffer();
    	sBuf.append("'" + candidate + "'" + separator);
    	for(int i=0,pos=0; i<N_GROUNDS; i++) {
    		String [][] cons = consequenceParams[i];
    		for(int j=0; j<cons.length; j++) {
        		sBuf.append("<M1" + grounds[i] + 
        				(groundParams[i]==null?"":Arrays.toString(groundParams[i]))
        				+ ",P" + getType(candidate.charAt(pos)) +
            			LAP + Arrays.toString(cons[j]) + ">" + separator);
        		pos++;
        		}
    		}
    	return sBuf.toString();
	}

	private static int getType(char c) {
		return consequenceType[c-'1'];
	}


	private double evaluateSituation() {
		if(wasteDALMASController.getDnrDALMASController().getCurrentSituation()==null)
			return 0;
		wasteDALMASController.refreshWastelandKB();
		wasteDALMASController.updateGrid();
		int collectedWaste = wasteDALMASController.getCollectedWaste();
		int uncollectedWaste = wasteDALMASController.getUncollectedWaste();
		double totalWaste = collectedWaste + uncollectedWaste;
    	return 100*collectedWaste/totalWaste;
//    	return collectedWaste;
	}


	/**
     * Always returns true.  A higher score indicates a fitter individual.
     * @return True.
     */
    public boolean isNatural()
    {
        return true;
    }
}
